from pydantic import BaseModel


class AnalysisSummary(BaseModel):
    order_id: int
    delivered: int | None = None
    running: int | None = None
    cancelled: int | None = None
    failed: int | None = None


class SummariesResponse(BaseModel):
    summaries: list[AnalysisSummary]



"""Module to read or write config files"""

from pathlib import Path
from typing import Any
from cg.constants.symbols import EMPTY_STRING


def write_config_nextflow_style(content: dict[str, Any] | None) -> str:
    """Write content to stream accepted by Nextflow config files with non-quoted booleans and quoted strings."""
    string: str = EMPTY_STRING
    double_quotes: str = '"'
    for parameter, value in content.items():
        if isinstance(value, Path):
            value: str = value.as_posix()
        quotes = double_quotes if type(value) is str else EMPTY_STRING
        string += f"params.{parameter} = {quotes}{value}{quotes}\n"
    return string



from pathlib import Path

import pytest

from cg.io.txt import read_txt, write_txt, concat_txt


def test_read_txt_to_list(txt_file_path: Path):
    """
    Test reading content from a TXT file into a list.
    """
    # GIVEN a txt file

    # WHEN reading the file
    content: list[str] = read_txt(file_path=txt_file_path)

    # THEN assert a list is returned
    assert isinstance(content, list)

    # THEN the content should match the expected lines
    expected_lines: list[str] = ["Line 1\n", "Line 2\n", "Line 3"]
    assert content == expected_lines


def test_read_txt_to_string(txt_file_path: Path):
    """
    Test reading content from a TXT file into a string.
    """
    # GIVEN a txt file

    # WHEN reading the file as a string
    content: str = read_txt(file_path=txt_file_path, read_to_string=True)

    # THEN assert a string is returned
    assert isinstance(content, str)

    # THEN the content should match the expected content
    expected_content: str = "Line 1\nLine 2\nLine 3"
    assert content == expected_content


def test_non_existent_file(non_existing_file_path: Path):
    """
    Test handling a non-existent file.
    """
    # GIVEN a path to a non-existing file

    # WHEN attempting to read a non-existent file
    with pytest.raises(FileNotFoundError):
        read_txt(non_existing_file_path)


def test_write_txt(txt_temp_path: Path, txt_file_path: Path):
    """Test writing content to a TXT file."""
    # GIVEN a txt file path to write to

    # GIVEN text content
    content: list[str] = read_txt(file_path=txt_file_path)

    # WHEN writing the file
    write_txt(content=content, file_path=txt_temp_path)

    # THEN assert file exists
    assert txt_temp_path.exists()

    # THEN the content should match the original content
    assert content == read_txt(file_path=txt_temp_path)


def test_concat_txt(txt_file_path: Path, txt_file_path_2: Path, txt_temp_path: Path):
    """Test concatenating two files, no optional string content"""
    # GIVEN a list of file paths to concatenate

    # WHEN concatenating two files
    concat_txt(
        file_paths=[txt_file_path, txt_file_path_2], target_file=txt_temp_path, str_content=None
    )

    # THEN the target file should exist
    assert txt_temp_path.exists()


def test_concat_txt_with_string(
    txt_file_path: Path, txt_file_path_2: Path, txt_temp_path: Path, csv_stream: str
):
    """Test concatenating two files, no optional string content"""
    # GIVEN a list of file paths to concatenate

    # WHEN concatenating two files
    concat_txt(
        file_paths=[txt_file_path, txt_file_path_2],
        target_file=txt_temp_path,
        str_content=[csv_stream],
    )

    # THEN the target file should exist
    assert txt_temp_path.exists()

    # THEN the content should match the input string
    assert read_txt(file_path=txt_temp_path) == [
        "Lorem,ipsum,sit,amet\n",
        "Line 1\n",
        "Line 2\n",
        "Line 3\n",
        "Line 4\n",
        "Line 5\n",
        "Line 6\n",
    ]



""" Test the environ app """

import getpass
import os

from cg.apps.environ import environ_email


def test_environ_email_sudo():
    """Test getting the email based on SUDO_USER"""
    # GIVEN $SUDO_USER is set to a diplomat
    sudo_user = "chrisjen.avasarala"
    os.environ["SUDO_USER"] = sudo_user

    # WHEN asking for the email
    email = environ_email()

    # THEN it should provide us with the email
    assert email == f"{sudo_user}@scilifelab.se"


def test_environ_email_env(monkeypatch):
    """Test getting the email based on logged in user"""
    # GIVEN $SUDO_USER is not set
    user = "chrisjen.avasarala"
    monkeypatch.setattr(getpass, "getuser", user)

    # WHEN asking for the email
    email = environ_email()

    # THEN it should provide us with the email
    assert email == f"{user}@scilifelab.se"



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import get_scout_link
from cg.store.models import Case


def get_case_message(case: Case) -> str:
    scout_link: str = get_scout_link(case)
    return f"Hello,\n\nThe following case has been uploaded to Scout:\n\n{scout_link}"


def get_cases_message(cases: list[Case]) -> str:
    scout_links: list[str] = [get_scout_link(case) for case in cases]
    scout_links_row_separated: str = "\n".join(scout_links)
    return (
        f"Hello,\n\nThe following cases has been uploaded to Scout:\n\n{scout_links_row_separated}"
    )


class ScoutMessage(DeliveryMessage):
    def create_message(self, cases: list[Case]) -> str:
        if len(cases) == 1:
            return get_case_message(cases[0])
        return get_cases_message(cases)



from datetime import datetime, timedelta
from typing import Generator

import pytest

from cg.constants import Workflow
from cg.constants.constants import CustomerId, PrepCategory
from cg.constants.subject import PhenotypeStatus
from cg.store.models import CaseSample, Order
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def cust123() -> str:
    """Return a customer id"""
    return "cust123"


@pytest.fixture
def test_subject() -> str:
    """Return a subject id"""
    return "test_subject"


@pytest.fixture
def store_with_samples_subject_id_and_tumour_status(
    store: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
) -> Store:
    """Return a store with two samples that have subject ids of which one is tumour"""
    helpers.add_sample(
        store=store,
        customer_id=cust123,
        is_tumour=True,
        internal_id="test_sample_1",
        name="sample_1",
        subject_id=test_subject,
    )

    helpers.add_sample(
        store=store,
        customer_id=cust123,
        is_tumour=False,
        internal_id="test_sample_2",
        name="sample_2",
        subject_id=test_subject,
    )
    return store


@pytest.fixture
def store_with_samples_and_tumour_status_missing_subject_id(
    store: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
) -> Store:
    """Return a store with two samples of which one is tumour"""
    helpers.add_sample(
        store=store,
        customer_id=cust123,
        is_tumour=True,
        internal_id="test_sample_1",
        name="sample_1",
    )

    helpers.add_sample(
        store=store,
        customer_id=cust123,
        is_tumour=False,
        internal_id="test_sample_2",
        name="sample_2",
    )
    return store


@pytest.fixture
def store_with_samples_customer_id_and_subject_id_and_tumour_status(
    store: Store, helpers: StoreHelpers
) -> Store:
    """Return a store with four samples with different customer IDs, and tumour status."""
    samples_data = [
        # customer_id, subject_id, is_tumour
        ("1", "test_subject", True),
        ("1", "test_subject_2", False),
        ("2", "test_subject", True),
        ("2", "test_subject_2", False),
    ]
    for customer_id, subject_id, is_tumour in samples_data:
        helpers.add_sample(
            store=store,
            customer_id=customer_id,
            is_tumour=is_tumour,
            internal_id=f"test_sample_{customer_id}_{subject_id}",
            name=f"sample_{customer_id}_{subject_id}",
            subject_id=subject_id,
        )
    return store


@pytest.fixture
def store_with_samples_that_have_names(store: Store, helpers: StoreHelpers) -> Store:
    """Return a store with two samples of which one has a name"""
    for index in range(1, 4):
        helpers.add_sample(
            store=store, internal_id=f"test_sample_{index}", name=f"test_sample_{index}"
        )

    helpers.add_sample(
        store=store,
        customer_id="unrelated_customer",
        internal_id="unrelated_id",
        name="unrelated_name",
    )
    return store


@pytest.fixture
def store_with_active_sample_analyze(
    store: Store, helpers: StoreHelpers
) -> Generator[Store, None, None]:
    """Return a store with an active sample with action analyze."""
    # GIVEN a store with a sample that is active
    case = helpers.add_case(
        store=store, name="test_case", internal_id="test_case_internal_id", action="analyze"
    )
    sample = helpers.add_sample(
        store=store, internal_id="test_sample_internal_id", name="test_sample"
    )
    helpers.add_relationship(store=store, sample=sample, case=case)

    yield store


@pytest.fixture
def store_with_active_sample_running(
    store: Store, helpers: StoreHelpers
) -> Generator[Store, None, None]:
    """Return a store with an active sample with action running."""
    # GIVEN a store with a sample that is active
    case = helpers.add_case(
        store=store, name="test_case", internal_id="test_case_internal_id", action="running"
    )
    sample = helpers.add_sample(
        store=store, internal_id="test_sample_internal_id", name="test_sample"
    )
    helpers.add_relationship(store=store, sample=sample, case=case)

    yield store


@pytest.fixture
def store_with_analyses_for_cases_not_uploaded_microsalt(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: datetime,
    timestamp_yesterday: datetime,
) -> Store:
    """Return a store with two analyses for two cases and workflow."""

    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            uploaded_at=timestamp_yesterday,
            delivery_reported_at=None,
            workflow=Workflow.MICROSALT,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            uploaded_at=None,
            delivery_reported_at=None,
            workflow=Workflow.MICROSALT,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
        link: CaseSample = analysis_store.relate_sample(
            case=oldest_analysis.case, sample=sample, status=PhenotypeStatus.UNKNOWN
        )
        analysis_store.session.add(link)
    return analysis_store


@pytest.fixture
def store_with_analyses_for_cases_to_deliver(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: datetime,
    timestamp_yesterday: datetime,
) -> Store:
    """Return a store with two analyses for two cases."""
    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            completed_at=timestamp_yesterday,
            uploaded_at=None,
            delivery_reported_at=None,
            workflow=Workflow.FLUFFY,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            completed_at=timestamp_now,
            uploaded_at=None,
            delivery_reported_at=None,
            workflow=Workflow.MIP_DNA,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=None)
        link: CaseSample = analysis_store.relate_sample(
            case=oldest_analysis.case, sample=sample, status=PhenotypeStatus.UNKNOWN
        )
        analysis_store.session.add(link)

    return analysis_store


@pytest.fixture
def store_with_multiple_pools_for_customer(
    store: Store,
    helpers: StoreHelpers,
    customer_id: str = CustomerId.CUST132,
) -> Generator[Store, None, None]:
    """Return a store with two pools with different names for the same customer."""
    for number in range(2):
        helpers.ensure_pool(
            store=store,
            customer_id=customer_id,
            name="_".join(["pool", str(number)]),
            order="_".join(["pool", "order", str(number)]),
        )
    yield store


@pytest.fixture
def re_sequenced_sample_store(
    store: Store,
    bcl_convert_flow_cell_id: str,
    case_id: str,
    family_name: str,
    bcl2fastq_flow_cell_id: str,
    sample_id: str,
    ticket_id: str,
    timestamp_now: datetime,
    helpers,
) -> Store:
    """Populate a store with a Fluffy case, with a sample that has been sequenced on two flow cells."""
    re_sequenced_sample_store: Store = store
    store_case = helpers.add_case(
        store=re_sequenced_sample_store,
        internal_id=case_id,
        name=family_name,
        data_analysis=Workflow.FLUFFY,
    )

    store_sample = helpers.add_sample(
        store=re_sequenced_sample_store,
        application_type=PrepCategory.READY_MADE_LIBRARY.value,
        is_tumour=False,
        internal_id=sample_id,
        reads=1200000000,
        original_ticket=ticket_id,
        last_sequenced_at=timestamp_now,
    )

    one_day_ahead_of_now = timestamp_now + timedelta(days=1)

    helpers.add_flow_cell(
        store=re_sequenced_sample_store,
        flow_cell_name=bcl_convert_flow_cell_id,
        samples=[store_sample],
        date=timestamp_now,
    )

    helpers.add_flow_cell(
        store=re_sequenced_sample_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        samples=[store_sample],
        date=one_day_ahead_of_now,
    )

    helpers.add_relationship(store=re_sequenced_sample_store, case=store_case, sample=store_sample)
    helpers.add_sample_lane_sequencing_metrics(
        store=re_sequenced_sample_store,
        sample_internal_id=store_sample.internal_id,
        flow_cell_name=bcl2fastq_flow_cell_id,
        flow_cell_lane_number=1,
        sample_total_reads_in_lane=120000000,
        sample_base_percentage_passing_q30=90,
    )
    return re_sequenced_sample_store


@pytest.fixture
def pool_name_1() -> str:
    """Return the name of the first pool."""
    return "pool_1"


@pytest.fixture
def pool_order_1() -> str:
    """Return the order of the first pool."""
    return "pool_order_1"


@pytest.fixture
def expected_number_of_not_archived_applications() -> int:
    """Return the number of expected numbers of not archived applications"""
    return 4


@pytest.fixture
def expected_number_of_applications() -> int:
    """Return the number of expected number of applications with prep category"""
    return 7


@pytest.fixture
def microbial_store(store: Store, helpers: StoreHelpers) -> Generator[Store, None, None]:
    """Populate a store with microbial application tags"""
    microbial_active_apptags = ["MWRNXTR003", "MWGNXTR003", "MWMNXTR003", "MWLNXTR003"]
    microbial_inactive_apptags = ["MWXNXTR003", "VWGNXTR001", "VWLNXTR001"]

    for app_tag in microbial_active_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="mic", is_archived=False)

    for app_tag in microbial_inactive_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="mic", is_archived=True)

    return store


@pytest.fixture
def max_nr_of_samples() -> int:
    """Return maximum numbers of samples"""
    return 50


@pytest.fixture
def max_nr_of_cases() -> int:
    """Return maximum numbers of cases"""
    return 50


@pytest.fixture
def three_customer_ids() -> list[str]:
    """Return three customer ids."""
    yield ["".join(["cust00", str(number)]) for number in range(3)]


@pytest.fixture
def store_with_pools_for_multiple_customers(
    store: Store, helpers: StoreHelpers, timestamp_now: datetime
) -> Generator[Store, None, None]:
    """Return a store with two samples for three different customers."""
    for number in range(3):
        helpers.ensure_pool(
            store=store,
            name="_".join(["test_pool", str(number)]),
            customer_id="".join(["cust00", str(number)]),
            no_invoice=False,
            delivered_at=timestamp_now,
        )
    yield store


@pytest.fixture
def three_pool_names() -> list[str]:
    """Return three customer ids."""
    yield ["_".join(["test_pool", str(number)]) for number in range(3)]


@pytest.fixture
def order(helpers: StoreHelpers, store: Store) -> Order:
    order: Order = helpers.add_order(
        store=store, customer_id=1, ticket_id=1, order_date=datetime.now()
    )
    return order


@pytest.fixture
def order_another(helpers: StoreHelpers, store: Store) -> Order:
    order: Order = helpers.add_order(
        store=store, customer_id=2, ticket_id=2, order_date=datetime.now()
    )
    return order


@pytest.fixture
def order_balsamic(helpers: StoreHelpers, store: Store) -> Order:
    order: Order = helpers.add_order(
        store=store,
        customer_id=2,
        ticket_id=3,
        order_date=datetime.now(),
        workflow=Workflow.BALSAMIC,
    )
    return order



from pathlib import Path

import pytest

from cg.constants.constants import Workflow
from cg.models.cg_config import CGConfig
from cg.store.models import Case


@pytest.fixture
def mutant_case(cg_context: CGConfig, case_id: str, ticket_id: str, helpers) -> Case:
    """Return mutant case"""
    case = helpers.add_case(
        store=cg_context.status_db,
        internal_id=case_id,
        name=ticket_id,
        data_analysis=Workflow.MUTANT,
    )
    return case


@pytest.fixture
def microsalt_case(cg_context: CGConfig, case_id: str, ticket_id: str, helpers) -> Case:
    """Return mutant case"""
    case = helpers.add_case(
        store=cg_context.status_db,
        internal_id=case_id,
        name=ticket_id,
        data_analysis=Workflow.MICROSALT,
    )
    return case


@pytest.fixture
def destination_path() -> Path:
    """Retyrbs a dummy path."""
    return Path("path", "to", "destination")



""" Test the CLI for run mip-dna """

import logging

import mock
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner
from pytest_mock import MockFixture

from cg.cli.workflow.mip_dna.base import start, start_available
from cg.constants.process import EXIT_SUCCESS
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from tests.cli.workflow.mip.conftest import setup_mocks
from tests.store.conftest import case_obj


def test_spring_decompression_needed_and_started(
    caplog: LogCaptureFixture,
    case: Case,
    cli_runner: CliRunner,
    mip_dna_context: CGConfig,
    mocker: MockFixture,
):
    """Tests starting the MIP analysis when decompression is needed"""
    caplog.set_level(logging.INFO)

    # GIVEN a case is available for analysis
    # GIVEN all samples in the case has dna application type
    # GIVEN the latest analysis has not started
    # GIVEN spring decompression is needed
    # GIVEN there are spring files that can be decompressed
    # GIVEN there are flow cells for the case
    setup_mocks(
        can_at_least_one_sample_be_decompressed=True,
        case_to_analyze=case,
        decompress_spring=True,
        has_latest_analysis_started=False,
        is_spring_decompression_needed=True,
        mocker=mocker,
    )

    # WHEN an MIP analysis is started
    result = cli_runner.invoke(start_available, obj=mip_dna_context)

    # THEN command should run without errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN it should be announced that spring decompression is started
    assert "Decompression started for" in caplog.text


def test_spring_decompression_needed_and_start_failed(
    caplog: LogCaptureFixture,
    case: Case,
    cli_runner: CliRunner,
    mip_dna_context: CGConfig,
    mocker: MockFixture,
):
    """Tests starting the MIP analysis when decompression is needed but fail to start"""
    caplog.set_level(logging.INFO)

    # GIVEN a case is available for analysis
    # GIVEN all samples in the case has dna application type
    # GIVEN the latest analysis has not started
    # GIVEN spring decompression is needed
    # GIVEN there are spring files that can be decompressed
    # GIVEN there are flow cells for the case
    setup_mocks(
        can_at_least_one_sample_be_decompressed=True,
        case_to_analyze=case,
        decompress_spring=False,
        has_latest_analysis_started=False,
        is_spring_decompression_needed=True,
        mocker=mocker,
    )

    # WHEN an MIP analysis is started
    result = cli_runner.invoke(start_available, obj=mip_dna_context)

    # THEN command should run without errors
    assert result.exit_code == 0

    # THEN it should be announced that spring decompression is needed but fail to start
    assert "Decompression failed to start for" in caplog.text


def test_spring_decompression_needed_and_cant_start(
    caplog: LogCaptureFixture,
    case: Case,
    cli_runner: CliRunner,
    mip_dna_context: CGConfig,
    mocker: MockFixture,
):
    """Tests starting the MIP analysis when decompression is needed but can't start"""
    caplog.set_level(logging.INFO)

    # GIVEN a case is available for analysis
    # GIVEN all samples in the case has dna application type
    # GIVEN the latest analysis has not started
    # GIVEN spring decompression is needed
    # GIVEN no spring files can be decompressed
    # GIVEN spring decompression is not running
    # GIVEN there are flow cells for the case
    setup_mocks(
        can_at_least_one_sample_be_decompressed=False,
        case_to_analyze=case,
        decompress_spring=False,
        has_latest_analysis_started=False,
        is_spring_decompression_needed=True,
        is_spring_decompression_running=False,
        mocker=mocker,
    )

    # WHEN an MIP analysis is started
    result = cli_runner.invoke(start_available, obj=mip_dna_context)

    # THEN command should run without errors
    assert result.exit_code == 0

    # THEN it should be announced that spring decompression is needed but fail to start
    assert "Decompression can not be started for" in caplog.text


def test_decompression_cant_start_and_is_running(
    mocker: MockFixture,
    cli_runner: CliRunner,
    caplog: LogCaptureFixture,
    mip_dna_context: CGConfig,
    case: Case,
):
    """Tests starting the MIP analysis when decompression is needed but can't start"""
    caplog.set_level(logging.INFO)

    # GIVEN a case is available for analysis
    # GIVEN all samples in the case has dna application type
    # GIVEN the latest analysis has not started
    # GIVEN spring decompression is needed
    # GIVEN no spring files can be decompressed
    # GIVEN spring decompression is running
    # GIVEN there are flow cells for the case
    setup_mocks(
        can_at_least_one_sample_be_decompressed=False,
        case_to_analyze=case,
        decompress_spring=False,
        has_latest_analysis_started=False,
        is_spring_decompression_needed=True,
        is_spring_decompression_running=True,
        mocker=mocker,
    )

    # WHEN an MIP analysis is started
    result = cli_runner.invoke(start_available, obj=mip_dna_context)

    # THEN command should run without errors
    assert result.exit_code == 0

    # THEN it should be announced that spring decompression is needed but fail to start
    assert "Decompression is running for" in caplog.text


def test_case_needs_to_be_stored(
    mocker: MockFixture,
    cli_runner: CliRunner,
    caplog: LogCaptureFixture,
    mip_dna_context: CGConfig,
    case: Case,
):
    """Test starting MIP when files are decompressed but not stored in housekeeper"""
    caplog.set_level(logging.INFO)

    # GIVEN a case is available for analysis
    # GIVEN all samples in the case has dna application type
    # GIVEN the latest analysis has not started
    # GIVEN spring decompression is not needed
    # GIVEN fastqs linked in housekeeper are checked successfully
    # GIVEN spring decompression is not running
    # GIVEN a panel file is created
    # GIVEN there are flow cells for the case
    setup_mocks(
        can_at_least_one_sample_be_decompressed=False,
        case_to_analyze=case,
        decompress_spring=False,
        has_latest_analysis_started=False,
        is_spring_decompression_needed=False,
        is_spring_decompression_running=False,
        mocker=mocker,
    )

    # GIVEN that, a panel is returned
    with mock.patch.object(
        mip_dna_context.scout_api,
        "export_panels",
        return_value=["OMIM-AUTO"],
    ):
        # WHEN MIP analysis is started
        result = cli_runner.invoke(start, [case.internal_id, "--dry-run"], obj=mip_dna_context)

    # THEN command should run without errors
    assert result.exit_code == 0

    # THEN the add_decompressed_fastq_files_to_housekeeper method should have been called
    assert PrepareFastqAPI.add_decompressed_fastq_files_to_housekeeper.call_count



"""Module for Rnafusion Analysis API."""

import logging
from pathlib import Path
from typing import Any

from cg import resources
from cg.constants import Workflow
from cg.constants.constants import FileFormat, Strandedness
from cg.constants.nf_analysis import MULTIQC_NEXFLOW_CONFIG, RNAFUSION_METRIC_CONDITIONS
from cg.resources import RNAFUSION_BUNDLE_FILENAMES_PATH
from cg.exc import MissingMetrics
from cg.io.controller import ReadFile
from cg.io.json import read_json
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase, MultiqcDataJson
from cg.models.fastq import FastqFileMeta
from cg.models.rnafusion.rnafusion import (
    RnafusionAnalysis,
    RnafusionParameters,
    RnafusionSampleSheetEntry,
)
from cg.store.models import Case, Sample

LOG = logging.getLogger(__name__)


class RnafusionAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RNAFUSION processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.RNAFUSION,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.rnafusion.root
        self.nfcore_workflow_path: str = config.rnafusion.workflow_path
        self.references: str = config.rnafusion.references
        self.profile: str = config.rnafusion.profile
        self.conda_env: str = config.rnafusion.conda_env
        self.conda_binary: str = config.rnafusion.conda_binary
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.rnafusion.tower_workflow
        self.account: str = config.rnafusion.slurm.account
        self.email: str = config.rnafusion.slurm.mail_user
        self.compute_env_base: str = config.rnafusion.compute_env
        self.revision: str = config.rnafusion.revision
        self.nextflow_binary_path: str = config.rnafusion.binary_path

    @property
    def use_read_count_threshold(self) -> bool:
        """Defines whether the threshold for adequate read count should be passed for all samples
        when determining if the analysis for a case should be automatically started."""
        return True

    def get_deliverables_template_content(self) -> list[dict]:
        """Return deliverables file template content."""
        return ReadFile.get_content_from_file(
            file_format=FileFormat.YAML,
            file_path=self.get_bundle_filenames_path(),
        )

    def get_nextflow_config_content(self) -> str:
        """Return nextflow config content."""
        return MULTIQC_NEXFLOW_CONFIG

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Rnafusion bundle filenames path."""
        return RNAFUSION_BUNDLE_FILENAMES_PATH

    def get_sample_sheet_content_per_sample(
        self, sample: Sample, case_id: str, strandedness: Strandedness
    ) -> list[list[str]]:
        """Get sample sheet content per sample."""
        sample_metadata: list[FastqFileMeta] = self.gather_file_metadata_for_sample(sample=sample)
        fastq_forward_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, forward_read=True
        )
        fastq_reverse_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, reverse_read=True
        )

        sample_sheet_entry = RnafusionSampleSheetEntry(
            name=case_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            strandedness=strandedness,
        )
        return sample_sheet_entry.reformat_sample_content()

    def get_sample_sheet_content(self, case_id: str, strandedness: Strandedness) -> list[list[Any]]:
        """Returns content for sample sheet."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        if len(case.links) != 1:
            raise NotImplementedError(
                "Case objects are assumed to be related to a single sample (one link)"
            )
        LOG.debug("Getting sample sheet information")
        for link in case.links:
            content_per_sample = self.get_sample_sheet_content_per_sample(
                sample=link.sample, case_id=case_id, strandedness=strandedness
            )
            return content_per_sample

    def get_workflow_parameters(
        self, case_id: str, genomes_base: Path | None = None
    ) -> RnafusionParameters:
        """Get Rnafusion parameters."""
        LOG.debug("Getting parameters information")
        return RnafusionParameters(
            cluster_options=f"--qos={self.get_slurm_qos_for_case(case_id=case_id)}",
            genomes_base=genomes_base or self.get_references_path(),
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
            priority=self.account,
        )

    def get_references_path(self, genomes_base: Path | None = None) -> Path:
        if genomes_base:
            return genomes_base.absolute()
        return Path(self.references).absolute()

    def config_case(
        self,
        case_id: str,
        strandedness: Strandedness,
        genomes_base: Path,
        dry_run: bool,
    ) -> None:
        """Create config files (parameters and sample sheet) for Rnafusion analysis."""
        self.create_case_directory(case_id=case_id, dry_run=dry_run)
        sample_sheet_content: list[list[Any]] = self.get_sample_sheet_content(
            case_id=case_id, strandedness=strandedness
        )
        workflow_parameters: RnafusionParameters = self.get_workflow_parameters(
            case_id=case_id, genomes_base=genomes_base
        )
        if dry_run:
            LOG.info("Dry run: Config files will not be written")
            return
        self.write_sample_sheet(
            content=sample_sheet_content,
            file_path=self.get_sample_sheet_path(case_id=case_id),
            header=RnafusionSampleSheetEntry.headers(),
        )
        self.write_params_file(case_id=case_id, workflow_parameters=workflow_parameters.dict())
        self.write_nextflow_config(case_id=case_id)

    def parse_multiqc_json_for_case(self, case_id: str) -> dict:
        """Parse a multiqc_data.json file and returns a dictionary with metric name and metric values for a case."""
        multiqc_json = MultiqcDataJson(
            **read_json(file_path=self.get_multiqc_json_path(case_id=case_id))
        )
        metrics_values: dict = {}
        for key in multiqc_json.report_general_stats_data:
            if case_id in key:
                metrics_values.update(list(key.values())[0])
                LOG.info(f"Key: {key}, Values: {list(key.values())[0]}")
        return metrics_values

    def get_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Get a multiqc_data.json file and returns metrics and values formatted."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_id: str = case.links[0].sample.internal_id
        metric_values: dict = self.parse_multiqc_json_for_case(case_id=case_id)
        metric_base_list: list = self.get_metric_base_list(
            sample_id=sample_id, metrics_values=metric_values
        )
        return metric_base_list

    @staticmethod
    def ensure_mandatory_metrics_present(metrics: list[MetricsBase]) -> None:
        """Check that all mandatory metrics are present. Raise error if missing."""
        given_metrics: set = {metric.name for metric in metrics}
        mandatory_metrics: set = set(RNAFUSION_METRIC_CONDITIONS.keys())
        LOG.info("Mandatory Metrics Keys:")
        for key in mandatory_metrics:
            LOG.info(key)
        missing_metrics: set = mandatory_metrics.difference(given_metrics)
        if missing_metrics:
            LOG.error(f"Some mandatory metrics are missing: {', '.join(missing_metrics)}")
            raise MissingMetrics()

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> RnafusionAnalysis:
        """Parse Rnafusion output analysis files and return analysis model."""
        sample_metrics: dict[str, dict] = {}
        for metric in qc_metrics_raw:
            try:
                sample_metrics[metric.id].update({metric.name.lower(): metric.value})
            except KeyError:
                sample_metrics[metric.id] = {metric.name.lower(): metric.value}
        return RnafusionAnalysis(sample_metrics=sample_metrics)

    def get_latest_metadata(self, case_id: str) -> RnafusionAnalysis:
        """Return the latest metadata of a specific Rnafusion case."""
        qc_metrics: list[MetricsBase] = self.get_multiqc_json_metrics(case_id)
        return self.parse_analysis(qc_metrics_raw=qc_metrics)

    def get_workflow_metrics(self) -> dict:
        return RNAFUSION_METRIC_CONDITIONS



from pathlib import Path

import pytest

from cg.constants.priority import SlurmQos
from cg.models.mip.mip_metrics_deliverables import MIPMetricsDeliverables


@pytest.fixture(name="mip_case_config_dna")
def mip_case_config_dna(fixtures_dir) -> Path:
    """Return path to MIP DNA case_config.yaml"""
    return Path(fixtures_dir, "apps", "mip", "dna", "store", "case_config.yaml")


@pytest.fixture(name="mip_analysis_config_dna_raw")
def mip_analysis_config_dna_raw() -> dict:
    """Raw MIP DNA analysis config"""
    return {
        "analysis_type": {"sample_id": "wgs"},
        "case_id": "yellowhog",
        "config_file_analysis": "a_config_path",
        "email": "someone@somewhere.com",
        "family_id": "a_family_id",
        "dry_run_all": False,
        "log_file": "a_log_path",
        "outdata_dir": "tests/fixtures/apps/mip/dna/store",
        "sample_ids": ["sample_id"],
        "sample_info_file": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
        "slurm_quality_of_service": SlurmQos.LOW,
        "store_file": "tests/fixtures/apps/mip/dna/store/case_id_deliverables.yaml",
    }


@pytest.fixture(name="mip_rank_model_version")
def mip_rank_model_version() -> str:
    """Return mip rank model version"""
    return "v1.0"


@pytest.fixture(name="mip_sv_rank_model_version")
def mip_sv_rank_model_version() -> str:
    """Return mip sv rank model version"""
    return "v1.2.0"


@pytest.fixture(name="sample_info_dna_raw")
def sample_info_dna_raw(mip_rank_model_version: str, mip_sv_rank_model_version: str) -> dict:
    """Raw sample_info fixture"""
    return {
        "analysisrunstatus": "finished",
        "analysis_date": "2021-05-05T16:16:01",
        "case_id": "yellowhog",
        "family_id": "a_family_id",
        "human_genome_build": {"version": 37, "source": "grch"},
        "mip_version": "v9.0.0",
        "recipe": {
            "genmod": {
                "rank_model": {"version": mip_rank_model_version},
            },
            "sv_genmod": {
                "sv_rank_model": {"version": mip_sv_rank_model_version},
            },
        },
    }


@pytest.fixture(name="mip_metrics_deliverables_raw")
def mip_metrics_deliverables_raw() -> dict:
    """Raw MIP metrics deliverables"""
    return {
        "metrics": [
            {
                "id": "an_id",
                "input": "file_1",
                "name": "raw_total_sequences",
                "step": "bamstats",
                "value": "296436536",
            },
            {
                "id": "an_id",
                "input": "file_2",
                "name": "raw_total_sequences",
                "step": "bamstats",
                "value": "287527252",
            },
            {
                "id": "an_id",
                "input": "file_1",
                "name": "reads_mapped",
                "step": "bamstats",
                "value": "295720636",
            },
            {
                "id": "an_id",
                "input": "file_2",
                "name": "reads_mapped",
                "step": "bamstats",
                "value": "286811812",
            },
            {
                "id": "another_id",
                "input": "file_1",
                "name": "raw_total_sequences",
                "step": "bamstats",
                "value": "286214720",
            },
            {
                "id": "another_id",
                "input": "file_2",
                "name": "raw_total_sequences",
                "step": "bamstats",
                "value": "284211662",
            },
            {
                "id": "another_id",
                "input": "file_1",
                "name": "reads_mapped",
                "step": "bamstats",
                "value": "285581150",
            },
            {
                "id": "another_id",
                "input": "file_2",
                "name": "reads_mapped",
                "step": "bamstats",
                "value": "283570947",
            },
            {
                "id": "an_id",
                "input": "some_input",
                "name": "MEAN_INSERT_SIZE",
                "step": "collectmultiplemetricsinsertsize",
                "value": "422.245916",
            },
            {
                "id": "another_id",
                "input": "some_input",
                "name": "MEAN_INSERT_SIZE",
                "step": "collectmultiplemetricsinsertsize",
                "value": "399.51625",
            },
            {
                "id": "an_id",
                "input": "some_input",
                "name": "MEDIAN_TARGET_COVERAGE",
                "step": "collecthsmetrics",
                "value": "29",
            },
            {
                "id": "another_id",
                "input": "some_input",
                "name": "MEDIAN_TARGET_COVERAGE",
                "step": "collecthsmetrics",
                "value": "28",
            },
            {
                "id": "an_id",
                "input": "some_input",
                "name": "gender",
                "step": "chanjo_sexcheck",
                "value": "female",
            },
            {
                "id": "another_id",
                "input": "some_input",
                "name": "gender",
                "step": "chanjo_sexcheck",
                "value": "male",
            },
            {
                "id": "an_id",
                "input": "some_input",
                "name": "fraction_duplicates",
                "step": "markduplicates",
                "value": "0.0748899652117993",
            },
            {
                "id": "another_id",
                "input": "some_input",
                "name": "fraction_duplicates",
                "step": "markduplicates",
                "value": "0.09",
            },
        ],
    }


@pytest.fixture(name="mip_metrics_deliverables")
def mip_metrics_deliverables(mip_metrics_deliverables_raw: dict) -> MIPMetricsDeliverables:
    return MIPMetricsDeliverables(**mip_metrics_deliverables_raw)


@pytest.fixture(name="mip_analysis_raw")
def mip_analysis_raw(
    case_id: str,
    mip_metrics_deliverables: MIPMetricsDeliverables,
    mip_rank_model_version: str,
    mip_sv_rank_model_version: str,
    sample_id,
) -> dict:
    return {
        "case": case_id,
        "genome_build": "grch37",
        "sample_id_metrics": mip_metrics_deliverables.sample_id_metrics,
        "mip_version": "10.0.0",
        "rank_model_version": mip_rank_model_version,
        "sample_ids": [sample_id],
        "sv_rank_model_version": mip_sv_rank_model_version,
    }



"""Mock models from the store. Used for processing them without using the store"""

import datetime

from cg.store.models import Analysis, Case, Customer


class Customer(Customer):
    """Mock a customer object"""

    def __init__(self):
        self.internal_id = "cust000"


class Case(Case):
    """Mock a case object"""

    def __init__(self):
        self.id: int = 1
        self.panels: list[str] = ["PEDHEP"]
        self.internal_id: str = "yellowhog"
        self.name: str = "analysis_family"
        self.customer: Customer = Customer()


class Analysis(Analysis):
    """Mock an store analysis object"""

    def __init__(self):
        self.id = 1
        self.completed_at = datetime.datetime.now()
        self.case: Case = Case()



"""Interactions with the gisaid cli upload_results_to_gisaid"""

import logging
import re
import tempfile
from pathlib import Path

import pandas as pd
from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.constants.constants import SARS_COV_REGEX, FileFormat
from cg.exc import HousekeeperFileMissingError
from cg.io.controller import ReadFile, WriteFile
from cg.models.cg_config import CGConfig
from cg.store.models import Sample
from cg.store.store import Store
from cg.utils import Process

from .constants import HEADERS
from .models import GisaidAccession, GisaidSample

LOG = logging.getLogger(__name__)

UPLOADED_REGEX_MATCH = r"\[\"([A-Za-z0-9_]+)\"\]\}$"


class GisaidAPI:
    """Interface with Gisaid cli uppload"""

    def __init__(self, config: CGConfig):
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.lims_api: LimsAPI = config.lims_api
        self.status_db: Store = config.status_db
        self.gisaid_submitter: str = config.gisaid.submitter
        self.upload_password: str = config.gisaid.upload_password
        self.upload_cid: str = config.gisaid.upload_cid
        self.gisaid_binary: str = config.gisaid.binary_path
        self.gisaid_log_dir: str = config.gisaid.log_dir
        self.log_watch: str = config.gisaid.logwatch_email
        self.email_base_settings = config.email_base_settings
        self.mutant_root_dir = Path(config.mutant.root)

        self.process = Process(binary=self.gisaid_binary)

    def get_completion_file_from_hk(self, case_id: str) -> File:
        """Find completon file in Housekeeper and return it"""

        completion_file: File | None = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["komplettering"]
        )
        if not completion_file:
            msg = f"completion file missing for bundle {case_id}"
            raise HousekeeperFileMissingError(message=msg)
        return completion_file

    def get_completion_dataframe(self, completion_file: File) -> pd.DataFrame:
        """Read completion file in to dataframe, drop duplicates, and return the dataframe"""
        completion_df = pd.read_csv(completion_file.full_path, index_col=None, header=0)
        completion_df.drop_duplicates(inplace=True)
        completion_df = completion_df[completion_df["provnummer"].str.contains(SARS_COV_REGEX)]
        return completion_df

    def get_gisaid_sample_list(self, case_id: str) -> list[Sample]:
        """Get list of Sample objects eligeble for upload.
        The criteria is that the sample reached 20x coverage for >95% bases.
        The sample will be included in completion file."""

        completion_file = self.get_completion_file_from_hk(case_id=case_id)
        completion_df = self.get_completion_dataframe(completion_file=completion_file)
        sample_names = list(completion_df["provnummer"].unique())
        return [self.status_db.get_sample_by_name(name=sample_name) for sample_name in sample_names]

    def get_gisaid_fasta_path(self, case_id: str) -> Path:
        """Get path to gisaid fasta"""
        return Path(self.mutant_root_dir, case_id, "results", f"{case_id}.fasta")

    def get_gisaid_csv_path(self, case_id: str) -> Path:
        """Get path to gisaid csv"""
        return Path(self.mutant_root_dir, case_id, "results", f"{case_id}.csv")

    def get_gisaid_samples(self, case_id: str) -> list[GisaidSample]:
        """Get list of Gisaid sample objects."""

        samples: list[Sample] = self.get_gisaid_sample_list(case_id=case_id)
        gisaid_samples = []
        for sample in samples:
            sample_id: str = sample.internal_id
            LOG.info(f"Creating GisaidSample for {sample_id}")
            gisaid_sample = GisaidSample(
                case_id=case_id,
                cg_lims_id=sample_id,
                covv_subm_sample_id=sample.name,
                submitter=self.gisaid_submitter,
                fn=f"{case_id}.fasta",
                covv_collection_date=str(
                    self.lims_api.get_sample_attribute(lims_id=sample_id, key="collection_date")
                ),
                region=self.lims_api.get_sample_attribute(lims_id=sample_id, key="region"),
                region_code=self.lims_api.get_sample_attribute(
                    lims_id=sample_id, key="region_code"
                ),
                covv_orig_lab=self.lims_api.get_sample_attribute(
                    lims_id=sample_id, key="original_lab"
                ),
                covv_orig_lab_addr=self.lims_api.get_sample_attribute(
                    lims_id=sample_id, key="original_lab_address"
                ),
            )
            gisaid_samples.append(gisaid_sample)
        return gisaid_samples

    def create_gisaid_fasta(self, gisaid_samples: list[GisaidSample], case_id: str) -> None:
        """Writing a new fasta with headers adjusted for gisaid upload_results_to_gisaid"""

        gisaid_fasta_file = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-fasta", case_id]
        )
        if gisaid_fasta_file:
            gisaid_fasta_path = gisaid_fasta_file.full_path
        else:
            gisaid_fasta_path: Path = self.get_gisaid_fasta_path(case_id=case_id)

        fasta_lines: list[str] = []

        for sample in gisaid_samples:
            fasta_file: File = self.housekeeper_api.get_file_from_latest_version(
                bundle_name=case_id, tags=[sample.cg_lims_id, "consensus-sample"]
            )
            if not fasta_file:
                raise HousekeeperFileMissingError(
                    message=f"No fasta file found for sample {sample.cg_lims_id}"
                )
            with open(str(fasta_file.full_path)) as handle:
                for line in handle.readlines():
                    if line[0] == ">":
                        fasta_lines.append(f">{sample.covv_virus_name}\n")
                    else:
                        fasta_lines.append(line)

        with open(gisaid_fasta_path, "w") as write_file_obj:
            write_file_obj.writelines(fasta_lines)

        if gisaid_fasta_file:
            return

        self.housekeeper_api.add_and_include_file_to_latest_version(
            bundle_name=case_id, file=gisaid_fasta_path, tags=["gisaid-fasta", case_id]
        )

    def create_gisaid_csv(self, gisaid_samples: list[GisaidSample], case_id: str) -> None:
        """Create csv file for gisaid upload"""
        samples_df = pd.DataFrame(
            data=[gisaid_sample.model_dump() for gisaid_sample in gisaid_samples],
            columns=HEADERS,
        )

        gisaid_csv_file = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-csv", case_id]
        )
        if gisaid_csv_file:
            LOG.info(f"GISAID CSV for case {case_id} exists, will be replaced")
            gisaid_csv_path = gisaid_csv_file.full_path
        else:
            gisaid_csv_path = self.get_gisaid_csv_path(case_id=case_id)
        samples_df.to_csv(gisaid_csv_path, sep=",", index=False)

        if gisaid_csv_file:
            return

        self.housekeeper_api.add_and_include_file_to_latest_version(
            bundle_name=case_id, file=gisaid_csv_path, tags=["gisaid-csv", case_id]
        )

    def create_gisaid_log_file(self, case_id: str) -> None:
        """Path for gisaid bundle log"""
        gisaid_log_file = self.housekeeper_api.get_files(
            bundle=case_id, tags=["gisaid-log", case_id]
        ).first()
        if gisaid_log_file:
            LOG.info("GISAID log exists in case bundle in Housekeeper")
            return

        log_file_path = Path(self.gisaid_log_dir, case_id).with_suffix(".log")
        if not log_file_path.parent.exists():
            raise ValueError(f"Gisaid log dir: {self.gisaid_log_dir} doesnt exist")
        if not log_file_path.exists():
            log_file_path.touch()
        self.housekeeper_api.add_and_include_file_to_latest_version(
            bundle_name=case_id, file=log_file_path, tags=["gisaid-log", case_id]
        )

    def create_gisaid_files_in_housekeeper(self, case_id: str) -> None:
        """Create all gisaid files in Housekeeper, if needed."""
        gisaid_samples = self.get_gisaid_samples(case_id=case_id)
        self.create_gisaid_csv(gisaid_samples=gisaid_samples, case_id=case_id)
        self.create_gisaid_fasta(gisaid_samples=gisaid_samples, case_id=case_id)
        self.create_gisaid_log_file(case_id=case_id)

    def authenticate_gisaid(self):
        load_call: list = [
            "CoV",
            "authenticate",
            "--cid",
            self.upload_cid,
            "--user",
            self.gisaid_submitter,
            "--pass",
            self.upload_password,
        ]
        self.process.run_command(parameters=load_call)

    def upload_results_to_gisaid(self, case_id: str) -> None:
        """Load batch data to GISAID using the gisiad cli."""

        temp_log_file = tempfile.NamedTemporaryFile(
            dir=self.gisaid_log_dir, mode="w+", delete=False
        )
        gisaid_csv_path = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-csv", case_id]
        ).full_path

        gisaid_fasta_path = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-fasta", case_id]
        ).full_path

        gisaid_log_path = (
            self.housekeeper_api.get_files(bundle=case_id, tags=["gisaid-log", case_id])
            .first()
            .full_path
        )

        self.authenticate_gisaid()
        load_call: list = [
            "--logfile",
            temp_log_file.name,
            "CoV",
            "upload",
            "--csv",
            gisaid_csv_path,
            "--fasta",
            gisaid_fasta_path,
        ]
        self.process.run_command(parameters=load_call)
        self.append_log(temp_log=Path(temp_log_file.name), gisaid_log=Path(gisaid_log_path))
        temp_log_file.close()
        if self.process.stderr:
            LOG.info(f"gisaid stderr:\n{self.process.stderr}")
        if self.process.stdout:
            LOG.info(f"gisaid stdout:\n{self.process.stdout}")

    def append_log(self, temp_log: Path, gisaid_log: Path) -> None:
        """Appends temp log to gisaid log and delete temp file"""
        new_log_data = ReadFile.get_content_from_file(
            file_format=FileFormat.JSON, file_path=temp_log.absolute()
        )
        if gisaid_log.stat().st_size != 0:
            old_log_data = ReadFile.get_content_from_file(
                file_format=FileFormat.JSON, file_path=gisaid_log.absolute()
            )
            new_log_data.extend(old_log_data)

        WriteFile.write_file_from_content(
            content=new_log_data, file_format=FileFormat.JSON, file_path=gisaid_log.absolute()
        )
        temp_log.unlink()

    def get_accession_numbers(self, case_id: str) -> dict[str, str]:
        """Parse accession numbers and sample ids from log file"""

        LOG.info("Parsing accession numbers from log file")
        accession_numbers = {}
        log_file = Path(
            self.housekeeper_api.get_files(bundle=case_id, tags=["gisaid-log", case_id])
            .first()
            .full_path
        )
        if log_file.stat().st_size != 0:
            log_data = ReadFile.get_content_from_file(
                file_format=FileFormat.JSON, file_path=log_file.absolute()
            )
            for log in log_data:
                if log.get("code") == "epi_isl_id":
                    log_message = log.get("msg")
                elif log.get("code") == "validation_error" and "existing_ids" in log.get("msg"):
                    log_message = (
                        f'{log.get("msg").split(";")[0]}; '
                        f'{re.findall(UPLOADED_REGEX_MATCH, log.get("msg"))[0]}'
                    )
                else:
                    continue
                accession_obj = GisaidAccession(log_message=log_message)
                accession_numbers[accession_obj.sample_id] = accession_obj.accession_nr
        return accession_numbers

    def update_completion_file(self, case_id: str) -> None:
        """Update completion file with accession numbers"""
        completion_file = self.get_completion_file_from_hk(case_id=case_id)
        accession_dict = self.get_accession_numbers(case_id=case_id)
        completion_df = self.get_completion_dataframe(completion_file=completion_file)
        completion_df["GISAID_accession"] = completion_df["provnummer"].apply(
            lambda x: accession_dict[x]
        )
        completion_df.to_csv(
            completion_file.full_path,
            sep=",",
            index=False,
        )

    def upload(self, case_id: str) -> None:
        """Uploading results to gisaid and saving the accession numbers in completion file"""

        completion_file = self.get_completion_file_from_hk(case_id=case_id)
        completion_df = self.get_completion_dataframe(completion_file=completion_file)
        if len(completion_df["GISAID_accession"].dropna()) == len(completion_df["provnummer"]):
            LOG.info("All samples already uploaded")
            return

        self.create_gisaid_files_in_housekeeper(case_id=case_id)
        self.upload_results_to_gisaid(case_id=case_id)
        self.update_completion_file(case_id=case_id)



"""API to prepare fastq files for analysis"""

import logging
from pathlib import Path

from housekeeper.store.models import File, Version

from cg.apps.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.constants import PIPELINES_USING_PARTIAL_ANALYSES
from cg.meta.compress import files
from cg.meta.compress.compress import CompressAPI
from cg.models import CompressionData
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class PrepareFastqAPI:
    def __init__(self, store: Store, compress_api: CompressAPI):
        self.store: Store = store
        self.hk_api: HousekeeperAPI = compress_api.hk_api
        self.compress_api: CompressAPI = compress_api
        self.crunchy_api: CrunchyAPI = compress_api.crunchy_api

    def get_compression_objects(self, case_id: str) -> list[CompressionData]:
        """Return a list of compression objects"""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        compression_objects = []
        for link in case.links:
            sample: Sample = link.sample
            sample_id = sample.internal_id
            if self._should_skip_sample(case=case, sample=sample):
                LOG.warning(f"Skipping sample {sample_id} - it has no reads.")
                continue
            version_obj: Version = self.compress_api.hk_api.get_latest_bundle_version(
                bundle_name=sample_id
            )
            compression_objects.extend(files.get_spring_paths(version_obj))
        return compression_objects

    def get_sample_compression_objects(self, sample_id: str) -> list[CompressionData]:
        compression_objects: list[CompressionData] = []
        version_obj: Version = self.compress_api.hk_api.get_latest_bundle_version(
            bundle_name=sample_id
        )
        compression_objects.extend(files.get_spring_paths(version_obj))
        return compression_objects

    def is_sample_decompression_needed(self, sample_id: str) -> bool:
        """Check if decompression is needed for the specified sample."""
        LOG.debug(f"Checking if decompression is needed for {sample_id}.")
        compression_objects = self.get_sample_compression_objects(sample_id=sample_id)
        return any(
            not self.crunchy_api.is_compression_pending(compression_object)
            and not compression_object.pair_exists()
            for compression_object in compression_objects
        )

    @staticmethod
    def _should_skip_sample(case: Case, sample: Sample):
        """
        For some workflows, we want to start a partial analysis disregarding the samples with no reads.
        This method returns true if we should skip the sample.
        """
        if case.data_analysis in PIPELINES_USING_PARTIAL_ANALYSES and not sample.has_reads:
            return True
        return False

    def is_spring_decompression_needed(self, case_id: str) -> bool:
        """Check if spring decompression needs to be started"""
        compression_objects = self.get_compression_objects(case_id=case_id)
        return any(
            not self.crunchy_api.is_compression_pending(compression_object)
            and not compression_object.pair_exists()
            for compression_object in compression_objects
        )

    def is_spring_decompression_running(self, case_id: str) -> bool:
        """Check if case is being decompressed"""
        compression_objects = self.get_compression_objects(case_id=case_id)
        return any(
            self.crunchy_api.is_compression_pending(compression_object)
            for compression_object in compression_objects
        )

    def can_at_least_one_sample_be_decompressed(self, case_id: str) -> bool:
        """Returns True if at least one sample can be decompressed, otherwise False"""
        compression_objects: list[CompressionData] = self.get_compression_objects(case_id=case_id)
        return any(
            self.crunchy_api.is_spring_decompression_possible(compression_object)
            for compression_object in compression_objects
        )

    def add_decompressed_fastq_files_to_housekeeper(self, case_id: str) -> None:
        """Adds decompressed FASTQ files to Housekeeper for a case, if there are any."""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        for link in case.links:
            sample: Sample = link.sample
            self.add_decompressed_sample(sample=sample, case=case)

    def add_decompressed_sample(self, sample: Sample, case: Case) -> None:
        """Adds decompressed FASTQ files to Housekeeper for a sample, if there are any."""
        sample_id = sample.internal_id
        if self._should_skip_sample(case=case, sample=sample):
            LOG.warning(f"Skipping sample {sample_id} - it has no reads.")
            return
        version: Version = self.compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
        fastq_files: dict[Path, File] = files.get_hk_files_dict(tags=["fastq"], version_obj=version)
        compressions: list[CompressionData] = files.get_spring_paths(version)
        for compression in compressions:
            self.add_decompressed_spring_object(
                compression=compression, fastq_files=fastq_files, sample=sample
            )

    def add_decompressed_spring_object(
        self, compression: CompressionData, fastq_files: dict[Path, File], sample: Case
    ) -> None:
        """Adds decompressed FASTQ files to Housekeeper related to a single spring file."""
        result = True
        sample_id: str = sample.internal_id
        if not self.are_fastqs_in_housekeeper(compression=compression, fastq_files=fastq_files):
            LOG.info(f"Adding FASTQ files to sample {sample_id} in Housekeeper")
            result: bool = self.compress_api.add_decompressed_fastq(sample)
        else:
            LOG.info(
                f"Both {compression.fastq_first} and {compression.fastq_second} "
                f"from sample {sample_id} are already in Housekeeper"
            )
        if not result:
            LOG.warning("Files were not added to fastq!")

    @staticmethod
    def are_fastqs_in_housekeeper(compression: CompressionData, fastq_files: dict[Path, File]):
        return compression.fastq_first in fastq_files and compression.fastq_second in fastq_files



"""Module for Raredisease Analysis API."""

import logging
from typing import Any
from pathlib import Path

from cg.io.txt import concat_txt
from cg.io.config import write_config_nextflow_style
from cg.constants import GenePanelMasterList, Workflow
from cg.constants.constants import FileExtensions
from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.meta.workflow.analysis import add_gene_panel_combo
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.fastq import FastqFileMeta
from cg.models.raredisease.raredisease import (
    RarediseaseSampleSheetEntry,
    RarediseaseSampleSheetHeaders,
)
from cg.models.nf_analysis import WorkflowParameters
from cg.store.models import Case, CaseSample


LOG = logging.getLogger(__name__)


class RarediseaseAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RAREDISEASE processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.RAREDISEASE,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.raredisease.root
        self.nfcore_workflow_path: str = config.raredisease.workflow_path
        self.references: str = config.raredisease.references
        self.profile: str = config.raredisease.profile
        self.conda_env: str = config.raredisease.conda_env
        self.conda_binary: str = config.raredisease.conda_binary
        self.config_platform: str = config.raredisease.config_platform
        self.config_params: str = config.raredisease.config_params
        self.config_resources: str = config.raredisease.config_resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.raredisease.tower_workflow
        self.account: str = config.raredisease.slurm.account
        self.compute_env: str = config.raredisease.compute_env
        self.revision: str = config.raredisease.revision

    def write_config_case(
        self,
        case_id: str,
        dry_run: bool,
    ) -> None:
        """Create a parameter (.config) files and a Nextflow sample sheet input for Raredisease analysis."""
        self.create_case_directory(case_id=case_id, dry_run=dry_run)
        sample_sheet_content: list[list[Any]] = self.get_sample_sheet_content(case_id=case_id)
        workflow_parameters: WorkflowParameters = self.get_workflow_parameters(case_id=case_id)
        if dry_run:
            LOG.info("Dry run: nextflow sample sheet and parameter file will not be written")
            return
        self.write_sample_sheet(
            content=sample_sheet_content,
            file_path=self.get_sample_sheet_path(case_id=case_id),
            header=RarediseaseSampleSheetHeaders.headers(),
        )
        self.write_params_file(case_id=case_id, workflow_parameters=workflow_parameters.dict())

    def get_sample_sheet_content_per_sample(
        self, case: Case = "", case_sample: CaseSample = ""
    ) -> list[list[str]]:
        """Get sample sheet content per sample."""
        sample_metadata: list[FastqFileMeta] = self.gather_file_metadata_for_sample(
            case_sample.sample
        )
        fastq_forward_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, forward_read=True
        )
        fastq_reverse_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, reverse_read=True
        )
        sample_sheet_entry = RarediseaseSampleSheetEntry(
            name=case_sample.sample.internal_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            sex=self.get_sex_code(case_sample.sample.sex),
            phenotype=self.get_phenotype_code(case_sample.status),
            paternal_id=case_sample.get_paternal_sample_id,
            maternal_id=case_sample.get_maternal_sample_id,
            case_id=case.internal_id,
        )
        return sample_sheet_entry.reformat_sample_content

    def get_sample_sheet_content(
        self,
        case_id: str,
    ) -> list[list[Any]]:
        """Return Raredisease nextflow sample sheet content for a case."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_sheet_content = []
        LOG.info("Getting sample sheet information")
        LOG.info(f"Samples linked to case {case_id}: {len(case.links)}")
        for link in case.links:
            sample_sheet_content.extend(
                self.get_sample_sheet_content_per_sample(case=case, case_sample=link)
            )
        return sample_sheet_content

    def get_workflow_parameters(self, case_id: str) -> WorkflowParameters:
        """Return parameters."""
        LOG.info("Getting parameters information")
        return WorkflowParameters(
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
        )

    def get_params_file_path(self, case_id: str, params_file: Path | None = None) -> Path:
        """Return parameters file or a path where the default parameters file for a case id should be located."""
        if params_file:
            return params_file.absolute()
        case_path: Path = self.get_case_path(case_id)
        return Path(case_path, f"{case_id}_params_file{FileExtensions.CONFIG}")
        # This function should be moved to nf-analysis to replace the current one when all nextflow pipelines are using the same config files approach

    def write_params_file(self, case_id: str, workflow_parameters: dict) -> None:
        """Write params-file for analysis."""
        LOG.debug("Writing parameters file")
        config_files_list = [self.config_platform, self.config_params, self.config_resources]
        extra_parameters_str = [
            write_config_nextflow_style(workflow_parameters),
            self.set_cluster_options(case_id=case_id),
        ]
        concat_txt(
            file_paths=config_files_list,
            target_file=self.get_params_file_path(case_id=case_id),
            str_content=extra_parameters_str,
        )

    @staticmethod
    def get_phenotype_code(phenotype: str) -> int:
        """Return Raredisease phenotype code."""
        LOG.debug("Translate phenotype to integer code")
        try:
            code = PlinkPhenotypeStatus[phenotype.upper()]
        except KeyError:
            raise ValueError(f"{phenotype} is not a valid phenotype")
        return code

    @staticmethod
    def get_sex_code(sex: str) -> int:
        """Return Raredisease sex code."""
        LOG.debug("Translate sex to integer code")
        try:
            code = PlinkSex[sex.upper()]
        except KeyError:
            raise ValueError(f"{sex} is not a valid sex")
        return code

    @property
    def root(self) -> str:
        return self.config.raredisease.root

    def write_managed_variants(self, case_id: str, content: list[str]) -> None:
        self._write_managed_variants(out_dir=Path(self.root, case_id), content=content)

    def write_panel(self, case_id: str, content: list[str]) -> None:
        """Write the gene panel to case dir."""
        self._write_panel(out_dir=Path(self.root, case_id), content=content)

    @staticmethod
    def get_aggregated_panels(customer_id: str, default_panels: set[str]) -> list[str]:
        """Check if customer should use the gene panel master list
        and if all default panels are included in the gene panel master list.
        If not, add gene panel combo and OMIM-AUTO.
        Return an aggregated gene panel."""
        master_list: list[str] = GenePanelMasterList.get_panel_names()
        if customer_id in GenePanelMasterList.collaborators() and default_panels.issubset(
            master_list
        ):
            return master_list
        all_panels: set[str] = add_gene_panel_combo(default_panels=default_panels)
        all_panels |= {GenePanelMasterList.OMIM_AUTO, GenePanelMasterList.PANELAPP_GREEN}
        return list(all_panels)

    def get_gene_panel(self, case_id: str) -> list[str]:
        """Create and return the aggregated gene panel file."""
        return self._get_gene_panel(case_id=case_id, genome_build=GENOME_BUILD_37)

    def get_managed_variants(self) -> list[str]:
        """Create and return the managed variants."""
        return self._get_managed_variants(genome_build=GENOME_BUILD_37)



"""Tests cli methods to create the case config for Raredisease."""

import logging

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.raredisease.base import config_case
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_config_case_dry_run(
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    caplog: LogCaptureFixture,
    raredisease_case_id: str,
):
    """Test dry-run."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a valid case

    # WHEN performing a dry-run
    result = cli_runner.invoke(config_case, [raredisease_case_id, "-d"], obj=raredisease_context)

    # THEN command should should exit successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN sample sheet and parameters information should be collected
    assert "Getting sample sheet information" in caplog.text
    assert "Getting parameters information" in caplog.text

    # THEN sample sheet and parameters information files should not be written

    assert "Dry run: nextflow sample sheet and parameter file will not be written" in caplog.text
    assert "Writing sample sheet" not in caplog.text
    assert "Writing parameters file" not in caplog.text


def test_config_case_without_samples(
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    caplog: LogCaptureFixture,
    no_sample_case_id: str,
):
    """Test config_case with a case without samples."""
    caplog.set_level(logging.ERROR)
    # GIVEN a case

    # WHEN running config case
    result = cli_runner.invoke(config_case, [no_sample_case_id], obj=raredisease_context)

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN warning should be printed that no sample is found
    assert no_sample_case_id in caplog.text
    assert "has no samples" in caplog.text


def test_config_case(
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    # rnafusion_sample_sheet_path: Path,
    # rnafusion_params_file_path: Path,
    caplog: LogCaptureFixture,
    raredisease_case_id: str,
):
    """Test case-config."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a valid case

    # WHEN performing a dry-run
    result = cli_runner.invoke(config_case, [raredisease_case_id], obj=raredisease_context)

    # THEN command should should exit successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN sample sheet and parameters information should be collected
    assert "Getting sample sheet information" in caplog.text
    assert "Getting parameters information" in caplog.text

    # THEN sample sheet and parameters information files should be written

    assert (
        "Dry run: nextflow sample sheet and parameter file will not be written" not in caplog.text
    )
    assert "Could not create config files" not in caplog.text
    assert "Writing sample sheet" in caplog.text
    assert "Writing parameters file" in caplog.text



"""CLI support to create config and/or start BALSAMIC PON analysis"""

import logging

import click

from cg.cli.workflow.balsamic.base import config_case, run, start
from cg.cli.workflow.commands import link, resolve_compression
from cg.meta.workflow.balsamic_pon import BalsamicPonAnalysisAPI

LOG = logging.getLogger(__name__)


@click.group("balsamic-pon", invoke_without_command=True)
@click.pass_context
def balsamic_pon(context: click.Context):
    """Cancer PON analysis workflow"""

    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return None

    config = context.obj
    context.obj.meta_apis["analysis_api"] = BalsamicPonAnalysisAPI(config=config)


balsamic_pon.add_command(resolve_compression)
balsamic_pon.add_command(link)
balsamic_pon.add_command(config_case)
balsamic_pon.add_command(run)
balsamic_pon.add_command(start)



from enum import Enum
from cg.apps.tb.dto.summary_response import AnalysisSummary
from cg.services.orders.order_status_service.dto.order_status_summary import OrderSummary
from cg.store.models import Case, Order


def create_summaries(
    orders: list[Order], analysis_summaries: list[AnalysisSummary]
) -> list[OrderSummary]:
    summaries: list[OrderSummary] = []
    for order in orders:
        summary: OrderSummary = create_order_summary(order)
        summaries.append(summary)
    add_analysis_summaries(order_summaries=summaries, analysis_summaries=analysis_summaries)
    return summaries


def add_analysis_summaries(
    order_summaries: list[OrderSummary],
    analysis_summaries: list[AnalysisSummary],
) -> None:
    order_summary_map = {summary.order_id: summary for summary in order_summaries}
    for analysis_summary in analysis_summaries:
        order_summary = order_summary_map[analysis_summary.order_id]
        order_summary.delivered = analysis_summary.delivered
        order_summary.running = analysis_summary.running
        order_summary.cancelled = analysis_summary.cancelled
        order_summary.failed = analysis_summary.failed


class CaseStatus(Enum):
    SEQUENCING = 1
    LAB_PREPARATION = 2
    OTHER = 3


def create_order_summary(order: Order) -> OrderSummary:
    in_sequencing: int = 0
    in_preparation: int = 0

    for case in order.cases:
        status: CaseStatus = get_case_status(case)
        if status == CaseStatus.SEQUENCING:
            in_sequencing += 1
        if status == CaseStatus.LAB_PREPARATION:
            in_preparation += 1

    return OrderSummary(
        order_id=order.id,
        total=len(order.cases),
        in_sequencing=in_sequencing,
        in_lab_preparation=in_preparation,
    )


def get_case_status(case: Case):
    """
    A case is in lab preparation if at least one sample is not prepared yet.
    A case is in sequencing if all samples have been prepared and at least one sample is in sequencing.
    """
    samples_in_sequencing = False
    for sample in case.samples:
        if sample.prepared_at is None:
            return CaseStatus.LAB_PREPARATION
        if sample.last_sequenced_at is None:
            samples_in_sequencing = True
    return CaseStatus.SEQUENCING if samples_in_sequencing else CaseStatus.OTHER



"""Tests CLI common methods to create the case config for NF analyses."""

import logging

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.base import workflow as workflow_cli
from cg.constants import EXIT_SUCCESS, Workflow
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_config_case_without_options(
    cli_runner: CliRunner, workflow: Workflow, request: FixtureRequest
):
    """Test config_case for workflow without options."""
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # WHEN dry running without anything specified
    result = cli_runner.invoke(workflow_cli, [workflow, "config-case"], obj=context)

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN command log should inform about missing arguments
    assert "Missing argument" in result.output


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_config_case_with_missing_case(
    cli_runner: CliRunner,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
    workflow: Workflow,
    request: FixtureRequest,
):
    """Test config_case for workflow with a missing case."""
    caplog.set_level(logging.ERROR)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case not in the StatusDB database
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id_does_not_exist)

    # WHEN running
    result = cli_runner.invoke(
        workflow_cli, [workflow, "config-case", case_id_does_not_exist], obj=context
    )

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN the error log should indicate that the case is invalid
    assert "could not be found in Status DB!" in caplog.text



from pathlib import Path

from pydantic.v1 import BaseModel, Field, conlist, validator

from cg.exc import SampleSheetError


class WorkflowParameters(BaseModel):
    input: Path
    outdir: Path


class NextflowSampleSheetEntry(BaseModel):
    """Nextflow sample sheet model.

    Attributes:
        name: sample name, or case id
        fastq_forward_read_paths: list of all fastq read1 file paths corresponding to sample
        fastq_reverse_read_paths: list of all fastq read2 file paths corresponding to sample
    """

    name: str
    fastq_forward_read_paths: conlist(Path, min_items=1)
    fastq_reverse_read_paths: conlist(Path, min_items=1)

    @validator("fastq_reverse_read_paths")
    def validate_complete_fastq_file_pairs(
        cls, fastq_reverse: list[str], values: dict
    ) -> list[str]:
        """Verify that the number of fastq forward files is the same as for the reverse."""
        if len(fastq_reverse) != len(values.get("fastq_forward_read_paths")):
            raise SampleSheetError("Fastq file length for forward and reverse do not match")
        return fastq_reverse

    @validator("fastq_forward_read_paths", "fastq_reverse_read_paths")
    def fastq_files_exist(cls, fastq_paths: list[str], values: dict) -> list[str]:
        """Verify that fastq files exist."""
        for fastq_path in fastq_paths:
            if not fastq_path.is_file():
                raise SampleSheetError(f"Fastq file does not exist: {str(fastq_path)}")
        return fastq_paths


class FileDeliverable(BaseModel):
    """Specification for a general deliverables file."""

    id: str
    format: str
    path: str
    path_index: str | None
    step: str
    tag: str

    @validator("path", "path_index", pre=True)
    def set_path_as_string(cls, file_path: str | Path) -> str | None:
        if file_path:
            return str(Path(file_path))
        return None


class WorkflowDeliverables(BaseModel):
    """Specification for workflow deliverables."""

    files: list[FileDeliverable]



import pytest

from cg.apps.demultiplex.sample_sheet.override_cycles_validator import OverrideCyclesValidator
from cg.constants.demultiplexing import FORWARD_INDEX_CYCLE_PATTERN, REVERSE_INDEX_CYCLE_PATTERN
from cg.exc import OverrideCyclesError


@pytest.mark.parametrize(
    "pattern, index_cycle, run_cycles, expected",
    [
        (FORWARD_INDEX_CYCLE_PATTERN, "I8N2", 10, True),
        (REVERSE_INDEX_CYCLE_PATTERN, "I8N2", 10, False),
        (FORWARD_INDEX_CYCLE_PATTERN, "N2I8", 10, False),
        (REVERSE_INDEX_CYCLE_PATTERN, "N2I8", 10, True),
        (FORWARD_INDEX_CYCLE_PATTERN, "NOT_A_CYCLE6N8", 10, False),
        (r"(\d+)", "I8N2", 10, False),
    ],
    ids=[
        "forward_valid",
        "reverse_invalid",
        "forward_invalid",
        "reverse_valid",
        "not_a_cycle",
        "not_a_pattern",
    ],
)
def test_is_index_cycle_value_following_pattern(
    override_cycles_validator: OverrideCyclesValidator,
    index2_8_nt_sequence_from_lims: str,
    pattern: str,
    index_cycle: str,
    run_cycles: int,
    expected: bool,
):
    """Test if an index follows a pattern works correctly."""
    # GIVEN an override cycles validator, an index sequence

    # WHEN checking if the index cycle value follows a pattern
    result: bool = override_cycles_validator._is_index_cycle_value_following_pattern(
        pattern=pattern,
        index_cycle=index_cycle,
        run_cycles=run_cycles,
        index_sequence=index2_8_nt_sequence_from_lims,
    )

    # THEN the result should be as expected
    assert result == expected


def test_validate_reads_cycles_correct_cycles(
    override_cycles_validator: OverrideCyclesValidator,
    forward_index2_cycle_processed_flow_cell_8_nt_sample: dict[str, str],
):
    """Test that a sample with correct read cycle values passes validation."""
    # GIVEN an override cycles validator with read cycles set
    override_cycles_validator.set_run_cycles(read1_cycles=151, read2_cycles=151)

    # GIVEN a sample with correct read cycle values
    override_cycles_validator.set_attributes_from_sample(
        forward_index2_cycle_processed_flow_cell_8_nt_sample
    )

    # WHEN validating the read cycles
    override_cycles_validator._validate_reads_cycles()

    # THEN no exception is raised


@pytest.mark.parametrize(
    "wrong_override_cycles",
    [
        "Y101;I8N2;I8N2;Y101",
        "Y151,I8N2,N2I8,Y151",
        "Y151;I8N2;",
        "NotAValidCycle",
    ],
    ids=["wrong_values", "wrong_separator", "missing_cycle", "not_a_valid_cycle"],
)
def test_validate_reads_cycles_incorrect_cycles(
    override_cycles_validator: OverrideCyclesValidator,
    processed_flow_cell_sample_8_index: dict[str, str],
    wrong_override_cycles: str,
    caplog: pytest.LogCaptureFixture,
):
    """Test that a sample with incorrect read cycle values raises an exception."""
    # GIVEN an override cycles validator read cycles set
    override_cycles_validator.set_run_cycles(read1_cycles=151, read2_cycles=151)

    # GIVEN a sample with incorrect read cycle values
    processed_flow_cell_sample_8_index["OverrideCycles"] = wrong_override_cycles
    override_cycles_validator.set_attributes_from_sample(processed_flow_cell_sample_8_index)

    # WHEN validating the read cycles
    with pytest.raises(OverrideCyclesError):
        override_cycles_validator._validate_reads_cycles()
    assert "Incorrect read cycles" in caplog.text


@pytest.mark.parametrize(
    "sample",
    [
        "forward_index2_cycle_processed_flow_cell_8_nt_sample",
        "reverse_index2_cycle_processed_flow_cell_8_nt_sample",
    ],
)
def test_validate_index1_cycles_correct_cycles(
    override_cycles_validator: OverrideCyclesValidator, sample: str, request: pytest.FixtureRequest
):
    """Test that a sample with correct index 1 cycle values passes validation."""
    # GIVEN an override cycles validator with index 1 cycles set
    override_cycles_validator.set_run_cycles(index1_cycles=10)

    # GIVEN a sample with correct index 1 cycle values
    override_cycles_validator.set_attributes_from_sample(request.getfixturevalue(sample))

    # WHEN validating the index 1 cycles
    override_cycles_validator._validate_index1_cycles()

    # THEN no exception is raised


@pytest.mark.parametrize(
    "run_index1_cycles,wrong_override_cycles",
    [
        (10, "Y151;N2I8;I8N2;Y151"),
        (10, "Y151;I10;I8N2;Y151"),
        (10, "Y151;I8;I8;Y151"),
        (10, "Y151;I8N9;I8;Y151"),
        (8, "Y151;I8N2;I8N2;Y151"),
        (8, "Y151;I10;N2I8;Y151"),
    ],
    ids=[
        "reverse_index1",
        "too_many_cycles",
        "missing_ignored_cycles",
        "custom_index_length",
        "extra_ignored_cycles",
        "extra_cycles",
    ],
)
def test_validate_index1_cycles_incorrect_cycles(
    override_cycles_validator: OverrideCyclesValidator,
    processed_flow_cell_sample_8_index: dict[str, str],
    run_index1_cycles: int,
    wrong_override_cycles: str,
    caplog: pytest.LogCaptureFixture,
):
    """Test that a sample with incorrect index 1 cycle values raises an exception."""
    # GIVEN an override cycles validator with index 1 cycles set
    override_cycles_validator.set_run_cycles(index1_cycles=run_index1_cycles)

    # GIVEN a sample with incorrect index 1 cycle values
    processed_flow_cell_sample_8_index["OverrideCycles"] = wrong_override_cycles
    override_cycles_validator.set_attributes_from_sample(processed_flow_cell_sample_8_index)

    # WHEN validating the index 1 cycles
    with pytest.raises(OverrideCyclesError):
        override_cycles_validator._validate_index1_cycles()
    assert "Incorrect index1 cycle" in caplog.text


@pytest.mark.parametrize(
    "sample, reverse_complement",
    [
        ("forward_index2_cycle_processed_flow_cell_8_nt_sample", False),
        ("reverse_index2_cycle_processed_flow_cell_8_nt_sample", True),
    ],
)
def test_validate_index2_cycles_correct_cycles(
    override_cycles_validator: OverrideCyclesValidator,
    sample: str,
    reverse_complement: bool,
    request: pytest.FixtureRequest,
):
    """Test that a sample with correct index 2 cycle values passes validation."""
    # GIVEN an override cycles validator with index 2 cycles and reverse complement set
    override_cycles_validator.set_run_cycles(index2_cycles=10)
    override_cycles_validator.set_reverse_complement(reverse_complement)

    # GIVEN a sample with correct index 2 cycle values
    override_cycles_validator.set_attributes_from_sample(request.getfixturevalue(sample))

    # WHEN validating the index 2 cycles
    override_cycles_validator._validate_index2_cycles()

    # THEN no exception is raised


@pytest.mark.parametrize(
    "run_index2_cycles,wrong_override_cycles,reverse_complement",
    [
        (10, "Y151;I8N2;I8N2;Y151", True),
        (10, "Y151;I8N2;N2I8;Y151", False),
        (10, "Y151;I8N2;I10;Y151", True),
        (10, "Y151;I8N2;I8;Y151", True),
        (10, "Y151;I8N2;I8N9;Y151", True),
        (8, "Y151;I8N2;I8N2;Y151", True),
        (8, "Y151;I8N2;I10;Y151", True),
    ],
    ids=[
        "forward_index2",
        "reverse_index2",
        "too_many_cycles",
        "missing_ignored_cycles",
        "custom_index_length",
        "extra_ignored_cycles",
        "extra_cycles",
    ],
)
def test_validate_index2_cycles_incorrect_cycles(
    override_cycles_validator: OverrideCyclesValidator,
    processed_flow_cell_sample_8_index: dict[str, str],
    run_index2_cycles: int,
    wrong_override_cycles: str,
    reverse_complement: bool,
    caplog: pytest.LogCaptureFixture,
):
    """Test that a sample with incorrect index 2 cycle values raises an exception."""
    # GIVEN an override cycles validator with index 2 cycles and reverse complement set
    override_cycles_validator.set_run_cycles(index2_cycles=run_index2_cycles)
    override_cycles_validator.set_reverse_complement(reverse_complement)

    # GIVEN a sample with incorrect index 2 cycle values
    processed_flow_cell_sample_8_index["OverrideCycles"] = wrong_override_cycles
    override_cycles_validator.set_attributes_from_sample(processed_flow_cell_sample_8_index)

    # WHEN validating the index 2 cycles
    with pytest.raises(OverrideCyclesError):
        override_cycles_validator._validate_index2_cycles()
    assert "does not match with run cycle" in caplog.text


@pytest.mark.parametrize(
    "sample, reverse_complement",
    [
        ("forward_index2_cycle_processed_flow_cell_8_nt_sample", False),
        ("reverse_index2_cycle_processed_flow_cell_8_nt_sample", True),
        ("processed_flow_cell_10_nt_sample", False),
    ],
    ids=["forward_8_nt", "reverse_8_nt", "10_nt"],
)
def test_validate_sample(
    override_cycles_validator: OverrideCyclesValidator,
    sample: str,
    reverse_complement: bool,
    request: pytest.FixtureRequest,
):
    """Test that a sample with correct override cycles passes validation."""
    # GIVEN a sample with correct override cycles

    # GIVEN an override cycles validator with all parameters set
    override_cycles_validator.set_run_cycles(
        read1_cycles=151, read2_cycles=151, index1_cycles=10, index2_cycles=10
    )
    override_cycles_validator.set_reverse_complement(reverse_complement)
    override_cycles_validator.set_attributes_from_sample(request.getfixturevalue(sample))

    # WHEN validating the sample
    override_cycles_validator.validate_sample(request.getfixturevalue(sample))

    # THEN no exception is raised



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import get_statina_link
from cg.store.models import Case


class StatinaMessage(DeliveryMessage):
    def create_message(self, cases: list[Case]) -> str:
        statina_link: str = get_statina_link(cases[0])
        return f"Hello,\n\nBatch {statina_link} is now available in Statina."



from pathlib import Path

import pytest


@pytest.fixture
def demultiplexed_bcl_convert_flow_cell(
    illumina_demultiplexed_runs_directory: Path, bcl_convert_flow_cell_full_name: str
) -> Path:
    return Path(illumina_demultiplexed_runs_directory, bcl_convert_flow_cell_full_name)



from pathlib import Path

from cg.io.gzip import read_gzip_first_line


def test_read_gzip_first_line(gzip_file_path: Path):
    """
    Test reading first line from a gzip file into a string.
    """
    # GIVEN a gzip file

    # WHEN reading the file
    line: str = read_gzip_first_line(file_path=gzip_file_path)

    # THEN assert a str is returned
    assert isinstance(line, str)

    # THEN the content should match the expected line
    assert line == "- ipsum, sit, amet"



"""Tests CLI common methods to store deliverable files into Housekeeper for NF analyses."""

import logging

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner
from pytest_mock import MockFixture

from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.hermes.models import CGDeliverables
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.workflow.base import workflow as workflow_cli
from cg.constants import EXIT_SUCCESS, Workflow
from cg.constants.constants import FileFormat
from cg.io.controller import WriteStream
from cg.models.cg_config import CGConfig
from cg.store.store import Store
from cg.utils import Process


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_store_housekeeper_without_options(
    cli_runner: CliRunner, workflow: Workflow, request: FixtureRequest
):
    """Test store-housekeeper for workflow without case id argument."""
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN no case id

    # WHEN invoking the command without additional parameters
    result = cli_runner.invoke(workflow_cli, [workflow, "store-housekeeper"], obj=context)

    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN command should mention argument
    assert "Missing argument" in result.output


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_store_housekeeper_with_missing_case(
    cli_runner: CliRunner,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
    request: FixtureRequest,
):
    """Test store-housekeeper for workflow with invalid case."""
    caplog.set_level(logging.ERROR)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case id not present in StatusDB
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id_does_not_exist)

    # WHEN running the store-housekeeper command
    result = cli_runner.invoke(
        workflow_cli, [workflow, "store-housekeeper", case_id_does_not_exist], obj=context
    )

    # THEN command should NOT succeed
    assert result.exit_code != EXIT_SUCCESS

    # THEN ERROR log should be printed containing the invalid case id
    assert case_id_does_not_exist in caplog.text
    assert "could not be found" in caplog.text


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_store_housekeeper_case_not_finished(
    cli_runner: CliRunner,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    request: FixtureRequest,
):
    """Test store-housekeeper for workflow using a case with an incomplete analysis."""
    caplog.set_level(logging.ERROR)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case id
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # WHEN running the store-housekeeper command
    result = cli_runner.invoke(workflow_cli, [workflow, "store-housekeeper", case_id], obj=context)

    # THEN command should NOT succeed
    assert result.exit_code != EXIT_SUCCESS

    # THEN warning should be printed that no deliverables file has been found
    assert "No deliverables file found for case" in caplog.text


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_store_housekeeper_case_with_malformed_deliverables_file(
    cli_runner,
    mocker,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    request: FixtureRequest,
):
    """Test store-housekeeper command for workflow using a case with malformed deliverables output."""
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")
    malformed_hermes_deliverables = request.getfixturevalue(
        f"{workflow}_malformed_hermes_deliverables"
    )

    # GIVEN that HermesAPI returns a malformed deliverables output
    mocker.patch.object(Process, "run_command")
    Process.run_command.return_value = WriteStream.write_stream_from_content(
        content=malformed_hermes_deliverables, file_format=FileFormat.JSON
    )

    # GIVEN a case id
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # WHEN running the store-housekeeper command
    result = cli_runner.invoke(workflow_cli, [workflow, "store-housekeeper", case_id], obj=context)

    # THEN command should NOT succeed
    assert result.exit_code != EXIT_SUCCESS
    assert "Could not store bundle in Housekeeper and StatusDB" in caplog.text


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_store_housekeeper_valid_case(
    cli_runner,
    workflow: Workflow,
    mocker,
    caplog: LogCaptureFixture,
    workflow_version: str,
    housekeeper_api: HousekeeperAPI,
    request: FixtureRequest,
):
    """Test store-housekeeper command for workflow with a valid case id."""
    caplog.set_level(logging.INFO)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")
    hermes_deliverables = request.getfixturevalue(f"{workflow}_hermes_deliverables")
    request.getfixturevalue(f"{workflow}_mock_deliverable_dir")
    store: Store = context.status_db

    # GIVEN a case id
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # GIVEN that case is not already stored in StatusDB
    assert not store.get_case_by_internal_id(internal_id=case_id).analyses

    # GIVEN that HermesAPI returns a deliverables output
    mocker.patch.object(HermesApi, "convert_deliverables")
    HermesApi.convert_deliverables.return_value = CGDeliverables(**hermes_deliverables)

    # WHEN running the store-housekeeper command
    result = cli_runner.invoke(workflow_cli, [workflow, "store-housekeeper", case_id], obj=context)

    # THEN a bundle should be successfully added to Housekeeper and StatusDB
    assert result.exit_code == EXIT_SUCCESS
    assert "Analysis successfully stored in Housekeeper" in caplog.text
    assert "Analysis successfully stored in StatusDB" in caplog.text
    assert store.get_case_by_internal_id(internal_id=case_id).analyses
    assert context.meta_apis["analysis_api"].housekeeper_api.bundle(case_id)

    # THEN a workflow version should be correctly stored in StatusDB
    assert (
        store.get_case_by_internal_id(internal_id=case_id).analyses[0].workflow_version
        == workflow_version
    )


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_valid_case_already_added(
    cli_runner,
    mocker,
    workflow: Workflow,
    real_housekeeper_api: HousekeeperAPI,
    caplog: LogCaptureFixture,
    request: FixtureRequest,
):
    """Test store-housekeeper command for workflow with a case already added in Housekeeper."""
    caplog.set_level(logging.INFO)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")
    hermes_deliverables = request.getfixturevalue(f"{workflow}_hermes_deliverables")
    request.getfixturevalue(f"{workflow}_mock_deliverable_dir")

    # GIVEN a case id
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # GIVEN a Housekeeper API
    context.housekeeper_api_: HousekeeperAPI = real_housekeeper_api
    context.meta_apis["analysis_api"].housekeeper_api = real_housekeeper_api

    # GIVEN that case is not already stored in StatusDB
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id).analyses

    # GIVEN that HermesAPI returns a deliverables output
    mocker.patch.object(HermesApi, "convert_deliverables")
    HermesApi.convert_deliverables.return_value = CGDeliverables(**hermes_deliverables)

    # GIVEN a case already stored in Housekeeper with existing bundles
    result_first = cli_runner.invoke(
        workflow_cli, [workflow, "store-housekeeper", case_id], obj=context
    )

    # GIVEN that the first command executed successfully
    assert result_first.exit_code == EXIT_SUCCESS

    # WHEN running the store-housekeeper command
    result = cli_runner.invoke(workflow_cli, [workflow, "store-housekeeper", case_id], obj=context)

    # THEN command should NOT succeed
    assert result.exit_code != EXIT_SUCCESS

    # THEN user should be informed that bundle was already added
    assert "Bundle already added" in caplog.text


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER],
)
def test_dry_run(
    cli_runner: CliRunner,
    workflow: Workflow,
    caplog: LogCaptureFixture,
    mocker: MockFixture,
    real_housekeeper_api: HousekeeperAPI,
    request: FixtureRequest,
):
    """Test dry run given a successfully finished case."""
    caplog.set_level(logging.INFO)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")
    hermes_deliverables = request.getfixturevalue(f"{workflow}_hermes_deliverables")
    request.getfixturevalue(f"{workflow}_mock_deliverable_dir")

    # GIVEN a case id
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    # GIVEN a Housekeeper API
    context.housekeeper_api_: HousekeeperAPI = real_housekeeper_api
    context.meta_apis["analysis_api"].housekeeper_api = real_housekeeper_api

    # GIVEN that HermesAPI returns a deliverables output
    mocker.patch.object(HermesApi, "convert_deliverables")
    HermesApi.convert_deliverables.return_value = CGDeliverables(**hermes_deliverables)

    # GIVEN that case is not already stored in Housekeeper
    assert not context.housekeeper_api.bundle(case_id)

    # GIVEN that case is not already stored in StatusDB
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id).analyses

    # WHEN running the store-housekeeper command with dry-run specified
    result = cli_runner.invoke(
        workflow_cli, [workflow, "store-housekeeper", case_id, "--dry-run"], obj=context
    )

    # THEN bundle should not be added to Housekeeper nor StatusDB
    assert result.exit_code == EXIT_SUCCESS
    assert "Dry-run: Housekeeper changes will not be commited" in caplog.text
    assert "Dry-run: StatusDB changes will not be commited" in caplog.text
    assert not context.housekeeper_api.bundle(case_id)



from pathlib import Path

import pytest

from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.apps.demultiplex.sample_sheet.sample_sheet_validator import SampleSheetValidator
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import BclConverter
from cg.io.controller import ReadFile


@pytest.fixture(scope="function")
def sample_sheet_validator() -> SampleSheetValidator:
    """Return a sample sheet validator."""
    return SampleSheetValidator()


@pytest.fixture
def hiseq_x_single_index_sample_sheet_content(
    hiseq_x_single_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a single-index HiSeq X sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_x_single_index_sample_sheet_path
    )


@pytest.fixture
def hiseq_x_single_index_bcl2fastq_sample_sheet_content(
    hiseq_x_single_index_bcl2fastq_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the path to a single-index Bcl2fastq HiSeq X sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_x_single_index_bcl2fastq_sample_sheet_path
    )


@pytest.fixture
def hiseq_x_dual_index_sample_sheet_content(
    hiseq_x_dual_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a dual-index HiSeq X sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_x_dual_index_sample_sheet_path
    )


@pytest.fixture
def hiseq_x_dual_index_bcl2fastq_sample_sheet_content(
    hiseq_x_dual_index_bcl2fastq_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the path to a dual-index Bcl2fastq HiSeq X sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_x_dual_index_bcl2fastq_sample_sheet_path
    )


@pytest.fixture
def hiseq_2500_dual_index_sample_sheet_content(
    hiseq_2500_dual_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a dual-index HiSeq 2500 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_2500_dual_index_sample_sheet_path
    )


@pytest.fixture
def hiseq_2500_dual_index_bcl2fastq_sample_sheet_content(
    hiseq_2500_dual_index_bcl2fastq_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the path to a dual-index Bcl2fastq HiSeq 2500 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_2500_dual_index_bcl2fastq_sample_sheet_path
    )


@pytest.fixture
def hiseq_2500_custom_index_sample_sheet_content(
    hiseq_2500_custom_index_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a custom-index HiSeq 2500 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_2500_custom_index_sample_sheet_path
    )


@pytest.fixture
def hiseq_2500_custom_index_bcl2fastq_sample_sheet_content(
    hiseq_2500_custom_index_bcl2fastq_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the path to a custom-index Bcl2fastq HiSeq 2500 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_2500_custom_index_bcl2fastq_sample_sheet_path
    )


@pytest.fixture
def novaseq_6000_pre_1_5_kits_sample_sheet_content(
    novaseq_6000_pre_1_5_kits_correct_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeq 6000 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_6000_pre_1_5_kits_correct_sample_sheet_path
    )


@pytest.fixture
def novaseq_6000_post_1_5_kits_sample_sheet_content(
    novaseq_6000_post_1_5_kits_correct_sample_sheet_path: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeq 6000 sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_6000_post_1_5_kits_correct_sample_sheet_path
    )


@pytest.fixture
def novaseq_x_sample_sheet_content(novaseq_x_correct_sample_sheet: Path) -> list[list[str]]:
    """Return the content of a NovaSeqX sample sheet."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_x_correct_sample_sheet
    )


@pytest.fixture
def novaseq_6000_sample_sheet_with_reversed_cycles_content(
    novaseq_6000_sample_sheet_with_reversed_cycles: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeq 6000 sample sheet with reversed cycles."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_6000_sample_sheet_with_reversed_cycles
    )


@pytest.fixture
def novaseq_x_sample_sheet_with_forward_cycles_content(
    novaseq_x_sample_sheet_with_forward_cycles: Path,
) -> list[list[str]]:
    """Return the content of a NovaSeqX sample sheet with forward cycles."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=novaseq_x_sample_sheet_with_forward_cycles
    )


@pytest.fixture
def sample_sheet_content_missing_data_header() -> list[list[str]]:
    """Return a sample sheet content with only headers."""
    return [["[Header]"], ["[Reads]"], ["[BCLConvert_Settings]"]]


@pytest.fixture
def novaseq_6000_post_1_5_kits_sample_sheet_object(
    sample_sheet_validator: SampleSheetValidator,
    novaseq_6000_post_1_5_kits_correct_sample_sheet_path: Path,
) -> SampleSheet:
    """Return a NovaSeq 6000 sample sheet object."""
    return sample_sheet_validator.get_sample_sheet_object_from_file(
        file_path=novaseq_6000_post_1_5_kits_correct_sample_sheet_path,
        bcl_converter=BclConverter.BCLCONVERT,
    )



from pathlib import Path

from cg.meta.workflow.microsalt.constants import QUALITY_REPORT_FILE_NAME
from cg.meta.workflow.microsalt.quality_controller import QualityController
from cg.meta.workflow.microsalt.quality_controller.models import QualityResult
from cg.models.cg_config import CGConfig
from cg.store.models import Application, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers

PRICES = {"standard": 1_000, "priority": 2_000, "express": 3_000, "research": 4_000}


def test_is_valid_total_reads_passes(quality_controller: QualityController):
    # GIVEN an application
    store = quality_controller.status_db
    application: Application = StoreHelpers.add_application(store=store, target_reads=1_000)

    # GIVEN an application version
    version = StoreHelpers.add_application_version(
        store=store,
        application=application,
        prices=PRICES,
    )

    # GIVEN a sample with a number of reads that is above the target reads
    sample: Sample = StoreHelpers.add_sample(store=store, reads=10_000)

    # GIVEN that the sample is associated with the application version
    sample.application_version = version

    # WHEN controlling the quality of the sample reads
    has_valid_reads: bool = quality_controller.has_valid_total_reads(sample.internal_id)

    # THEN the sample passes the quality control
    assert has_valid_reads


def test_is_valid_total_reads_fails(quality_controller: QualityController):
    # GIVEN an application
    store = quality_controller.status_db
    application: Application = StoreHelpers.add_application(store=store, target_reads=1_000)

    # GIVEN an application version
    version = StoreHelpers.add_application_version(
        store=store,
        application=application,
        prices=PRICES,
    )

    # GIVEN a sample with a number of reads that is far below the target reads
    sample: Sample = StoreHelpers.add_sample(store=store, reads=100)

    # GIVEN that the sample is associated with the application version
    sample.application_version = version

    # WHEN controlling the quality of the sample reads
    has_valid_reads: bool = quality_controller.has_valid_total_reads(sample.internal_id)

    # THEN the sample fails the quality control
    assert not has_valid_reads


def test_quality_control_fails(qc_microsalt_context: CGConfig, metrics_file_failing_qc: Path):
    # GIVEN a metrics file with samples that should fail the quality control

    # GIVEN a store containing the corresponding samples
    store: Store = qc_microsalt_context.status_db

    # GIVEN a quality controller
    quality_controller = QualityController(store)

    # WHEN performing the quality control
    result: QualityResult = quality_controller.quality_control(metrics_file_failing_qc)

    # THEN the case should fail the quality control
    assert not result.passes_qc

    # THEN a report should be generated
    assert metrics_file_failing_qc.parent.joinpath(QUALITY_REPORT_FILE_NAME).exists()


def test_quality_control_passes(qc_microsalt_context: CGConfig, metrics_file_passing_qc: Path):
    # GIVEN a metrics file with samples that should pass the quality control

    # GIVEN a store containing the corresponding samples
    store: Store = qc_microsalt_context.status_db

    # GIVEN a quality controller
    quality_controller = QualityController(store)

    # WHEN performing the quality control
    result: QualityResult = quality_controller.quality_control(metrics_file_passing_qc)

    # THEN the case should pass the quality control
    assert result.passes_qc

    # THEN a report should be generated
    assert metrics_file_passing_qc.parent.joinpath(QUALITY_REPORT_FILE_NAME).exists()



from pydantic import BaseModel, Field


class OrderSummary(BaseModel):
    order_id: int = Field(exclude=True)
    total: int
    delivered: int | None = None
    running: int | None = None
    cancelled: int | None = None
    failed: int | None = None
    in_sequencing: int
    in_lab_preparation: int
